"use strict";

var _            = require("lodash");
var Immutable    = require("immutable");

var utils        = require("./utils");
var connectUtils = require("./connect-utils");

module.exports = {
    /**
     * @param bs
     * @param done
     */
    getEmptyPort: function (bs, done) {
        utils.getPorts(bs.options, function (err, port) {
            if (err) {
                return utils.fail(true, err, bs.cb);
            }
            bs.debug("Found a free port: {magenta:%s", port);
            done(null, {
                options: {
                    port: port
                }
            });
        });
    },
    /**
     * @param {BrowserSync} bs
     * @param {Function} done
     */
    getOnlineStatus: function (bs, done) {
        if (_.isUndefined(bs.options.get("online")) && _.isUndefined(process.env.TESTING)) {
            require("dns").resolve("www.google.com", function (err) {
                var online = false;
                if (err) {
                    bs.debug("Could not resolve www.google.com, setting {magenta:online: false}");
                } else {
                    bs.debug("Resolved www.google.com, setting {magenta:online: true}");
                    online = true;
                }
                done(null, {
                    options: {
                        online: online
                    }
                });
            });
        } else {
            done();
        }
    },
    /**
     * @param bs
     * @param done
     */
    setOptions: function (bs, done) {
        done(null, {
            options: {
                urls:        utils.getUrlOptions(bs.options),
                files:       bs.pluginManager.hook("files:watch", bs.options.get("files"), bs.pluginManager.pluginOptions),
                snippet:     connectUtils.scriptTags(bs.options),
                scriptPaths: Immutable.fromJS(connectUtils.clientScript(bs.options, true)),
                mode:        (function () {
                    if (bs.options.get("server")) {
                        return "server";
                    }
                    if (bs.options.get("proxy")) {
                        return "proxy";
                    }
                    return "snippet";
                })()
            }
        });
    },
    /**
     * @param {BrowserSync} bs
     * @param {Function} done
     */
    setInternalEvents: function (bs, done) {
        require("./internal-events")(bs);
        done();
    },
    /**
     * @param {BrowserSync} bs
     * @param {Function} done
     */
    setFileWatchers: function (bs, done) {
        done(null, {
            instance: {
                watchers: bs.pluginManager.get("file:watcher")(bs.options, bs.events)
            }
        });
    },
    /**
     * @param {BrowserSync} bs
     * @param {Function} done
     */
    startServer: function (bs, done) {

        var clientJs = bs.pluginManager.hook("client:js", {
            port: bs.options.get("port"),
            options: bs.options
        });

        var server = bs.pluginManager.get("server")(bs,
            bs.pluginManager.get("client:script")(bs.options.toJS(), clientJs, bs.options.get("proxy") ? "file" : "middleware")
        );

        done(null, {
            instance: {
                clientJs: clientJs,
                server: server.server,
                app: server.app
            }
        });
    },
    /**
     * @param {BrowserSync} bs
     * @param {Function} done
     */
    startTunnel: function (bs, done) {
        if (bs.options.get("tunnel") && bs.options.get("online")) {
            var localTunnel = require("./tunnel");
            localTunnel(bs, function (err, tunnel) {
                if (err) {
                    return done(err);
                } else {
                    return done(null, {
                        optionsIn: [
                            {
                                path:  ["urls", "tunnel"],
                                value: tunnel.url
                            }
                        ],
                        instance: {
                            tunnel: tunnel
                        }
                    });
                }
            });
        } else {
            done();
        }
    },
    /**
     * @param {BrowserSync} bs
     * @param {Function} done
     */
    startSockets: function (bs, done) {

        var clientEvents = bs.pluginManager.hook("client:events", bs.clientEvents); // todo - Immutable?

        // Start the socket, needs an existing server.
        var io = bs.pluginManager.get("socket")(
            bs.server,
            clientEvents,
            bs
        );

        done(null, {
            instance: {
                clientEvents: clientEvents,
                io: io
            }
        });
    },
    /**
     *
     * @param bs
     * @param done
     */
    startUi: function (bs, done) {

        var userPlugins = bs.getUserPlugins();
        var ui          = bs.pluginManager.get("UI");
        var uiOpts;

        // if user provided a UI, use it instead
        if (userPlugins.some(function (item) {
            return item.name === "UI";
        })) { // do nothing if user-provided plugin
            delete bs.pluginManager.defaultPlugins["UI"];
            return done();
        }

        if (!(uiOpts = bs.options.get("ui"))) {
            return done();
        }

        var opts      = uiOpts.toJS();
        opts.logLevel = bs.options.get("logLevel");

        if (ui) {
            ui = ui(opts, bs);
            return ui.getServer(function (err, server) {
                done(null, {
                    instance: {
                        ui: {
                            server: server,
                            instance: ui
                        }
                    }
                });
            });
        }

        return done();
    },
    /**
     * @param {BrowserSync} bs
     * @param {Function} done
     */
    initUserPlugins: function (bs, done) {

        bs.pluginManager.initUserPlugins(bs);

        done(null, {
            options: {
                userPlugins: bs.getUserPlugins()
            }
        });
    }
};
